/* 
 * brl-fcgi.c --
 *
 *	BRL-BUDDY helps scheme produce a page containing BRL output
 *
 */

#define BRL_DIR_DEFAULT "/usr/local/brl"

static const char rcsid[] = "$Id: brlfcgi.c,v 1.12 1997/10/02 20:23:10 brlewis Exp $";

#include "config.h"
#ifdef HAVE_LIBFCGI
extern char **environ;
static char **environ_initial;
/* DO NOT #include <stdio.h> */
#include <fcgi_stdio.h>
#else
#include <stdio.h>
#endif
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <signal.h>
#include <string.h>
#include <sys/types.h>
#include <errno.h>
#include <sys/wait.h>

extern char **environ;

static int brl_catches[] = { SIGABRT, SIGALRM, SIGCHLD, SIGHUP, SIGINT,
			     SIGPIPE, SIGQUIT, SIGSEGV, SIGTERM, 0 };

static char scm_nomem[]="(error \"Out of memory\")\n";
static char hexdigits[]="0123456789ABCDEF";
static pid_t scheme_pid=0;
static int client_connected=0;
#ifndef NO_BRL_DEBUG
int debugfd;
#endif

void
brl_die(signo)
     int signo;
{
  if (signo) fprintf(stderr, "signal %d...exiting\n", signo);
  fflush(stderr);
  if (scheme_pid) kill(scheme_pid, SIGTERM);
  fputs("Content-type: text/html\r\n\r\n", stdout);
  fputs("<h1>Service temporarily unavailable. Try again.</h1>\n", stdout);
  fputs("Click the Reload button on your browser (Ctrl-R in lynx).\n", stdout);
#ifdef HAVE_LIBFCGI
  FCGI_Finish();
#endif
  exit(1);
}

/*
 * Poor man's buffered I/O.  The fromfile is buffered, the tofile
 * presumably can't be buffered (input_to_scheme).
 */

int
brl_flush(fromfile, tofile)
     FILE *fromfile, *tofile;
{
  char buf[BUFSIZ];
  int nbytes;

  if (ftell(fromfile) > 0)
    {
      fflush(fromfile);
      rewind(fromfile);
      do
	{
	  if (0 != (nbytes = fread(buf, sizeof(char), BUFSIZ, fromfile)))
	    {
	      if (!fwrite(buf, sizeof(char), nbytes, tofile)) 
		break;
#ifndef NO_BRL_DEBUG
		if (debugfd != -1) write(debugfd, buf, nbytes);
#endif
	    }
	} while (nbytes == BUFSIZ);
      fflush(tofile);
      rewind(fromfile);
      ftruncate(fileno(fromfile), 0);
    }
  return ferror(tofile);
}

/* Write a count of characters as part of a Scheme string */

void
brl_substr2subsexp(scheme_file, str, len)
     FILE *scheme_file;
     char *str;
     int len;
{
  char *ptr1, *ptr2, ch3[3];

  strcpy(ch3, "\\ ");
  for(ptr1=str; ptr1 < str + len; )
    {
      /* Find quote or backslash */
      ptr2=strpbrk(ptr1, "\"\\");
      if (!ptr2 || ptr2 >= str + len)
	{
	  fwrite(ptr1, 1, len - (ptr1 - str), scheme_file);
	  break;
	}
      ch3[1]=*ptr2;
      fwrite(ptr1, 1, ptr2 - ptr1, scheme_file);
      fwrite(ch3, 1, 2, scheme_file);
      ptr1=ptr2+1;
    }

  return;
}

/* Write a C string as a scheme expr.  Length of -1 means use strlen */

void
brl_str2sexp(scheme_file, str, len)
     FILE *scheme_file;
     char *str;
     int len;
{
  if (len == -1) len = strlen(str);
  fwrite("\"", 1, 1, scheme_file);
  brl_substr2subsexp(scheme_file, str, len);
  fwrite("\"", 1, 1, scheme_file);
  return;
}

/*
 * Non-alphanumeric chars that won't confuse the scheme CGI request handler
 * See also: http://www.w3.org/pub/WWW/Addressing/URL/5_BNF.html
 */
static char url_safe[]="/:$-_@.";

/* Write contents of the environment to fp as a scheme expression */

void
environ2sexp(fp)
     FILE *fp;
{
  char **env_var, *ptr1, *ptr2;

  fputs("(\n", fp);

  for(env_var=environ; *env_var; env_var++)
    {
      ptr1=*env_var;
      fputs("(", fp);

      /* Print out the NAME= part */
      ptr2=strchr(ptr1, '=');
      if (!ptr2)
	{
	  /* No =value part.  Print null value */
	  fprintf(fp, "%s ())\n", *env_var);
	  continue;
	}
      else
	{
	  /* assume env var is good scheme sym */
	  fwrite(*env_var, 1, ptr2-ptr1, fp);
	  /* Write string value */
	  fputs(" ", fp);
	  brl_str2sexp(fp, ptr2+1, -1);
	  fputs(")\n", fp);
	}
    }
  fputs(")\n", fp);
  return;
}

/*
 * See if a character is ok as 1st char in scheme sym.
 * The percent sign is deliberately omitted - reserved for encoding invalid
 * characters.
 */
int
brl_scmsym1(ch)
     int ch;
{
  return (int)(isalpha(ch) || strchr(".*/<=>!?:$_&~^", ch));
}

/*
 * See if a character is ok as non-1st char in scheme sym.
 * The percent sign is deliberately omitted - reserved for encoding invalid
 * characters.
 */
int
brl_scmsym2(ch)
     int ch;
{
  return (int)(isalnum(ch) || strchr("+-.*/<=>!?:$_&~^", ch));
}

static char *query_string=(char *)0;
static int content_length=0;
static int get_method=0;

int
next_query_char()
{
  if (content_length<=0) return EOF;

  content_length--;
  if (get_method)
    return (int) (*query_string++);
  else
    return getc(stdin);
}

static char lastchars[4];

int
next_translated_query_char()
{
  int ch;

  switch (ch=next_query_char())
    {
    case EOF:			/* no more chars in stream */
      return EOF;
      break;
    case (int)'+':
      strcpy(lastchars, "%20");
      return (int)' ';
      break;
    case (int)'%':
      lastchars[0]=(char)ch;
      if (EOF == (ch=next_query_char())) return EOF;
      else lastchars[1]=(char)ch;
      if (EOF == (ch=next_query_char())) return EOF;
      else lastchars[2]=(char)ch;
      lastchars[3]='\0';
      sscanf(lastchars+1, "%x", &ch);
      return ch;
      break;
    default:
      lastchars[0]=(char)ch;
      return ch;
      break;
    }
  /*NOTREACHED*/
  return ch;
}

/* Translate an HTML form input name into a scheme symbol */
int
iname2sym(fp)
     FILE *fp;
{
  int ch, firstchar=1;

  while ( (ch=next_translated_query_char()) != EOF
	  && lastchars[0] != (int)'=' )
    {
      if (firstchar) fputc('(', fp);
      if (firstchar ? brl_scmsym1(ch) : brl_scmsym2(ch)) fputc(ch, fp);
      else fputs(lastchars, fp);
      firstchar=0;
    }
  return ch;
}

/* Translate an HTML form input value into a scheme string */
int
ival2str(fp)
     FILE *fp;
{
  int ch;

  fputs(" \"", fp);
  while ( (ch=next_translated_query_char()) != EOF
	  && !strchr("&;", lastchars[0]) )
    {
      if (strchr("\"\\", ch)) fputc((int)'\\', fp);
      fputc(ch, fp);
    }
  fputs("\")\n", fp);
  return ch;
}

void
query2sexp(fp)
     FILE *fp;
{
  char *env_var;

  env_var=getenv("REQUEST_METHOD");
  if (!env_var)
    {
      printf("REQUEST_METHOD not set.");
      return;
    }

  get_method = !strcmp(env_var, "GET");
  if (get_method)
    {
      query_string=getenv("QUERY_STRING");
      content_length=strlen(query_string);
    }
  else
    content_length = atoi(getenv("CONTENT_LENGTH"));

  /* Start converting the query string to a scheme expression */
  fputs("(\n", fp);
  while (1)
    {
      if (EOF == iname2sym(fp)) break;
      if (EOF == ival2str(fp)) break;
    }
  fputs(")\n", fp);
  return;
}

/* Provides a 2-way pipe to scheme.  Returns 0 on success. */

int
scheme_pipe(filedes)
     int filedes[];		/* function sets [0] for read, [1] for write */
{
  int pipe_parent_to_child[2];
  int pipe_child_to_parent[2];
  pid_t pid_my_child;
  char *env_var, pathbuf[1024];

  /* Open the pipes */
  if (pipe(pipe_parent_to_child) < 0 || pipe(pipe_child_to_parent) < 0)
    return -1;

  /* fork */
  if ((pid_my_child = fork()) < 0)
    return -1;
  else if (pid_my_child == 0)
    {
      /* I am the child process.  Change my file descriptors. */
      dup2(pipe_parent_to_child[0], 0);
      dup2(pipe_child_to_parent[1], 1);
      dup2(pipe_child_to_parent[1], 2);	/* catch scheme's error output */

      /* Exec the program */
      env_var = getenv("MITSCHEME_LIBRARY_PATH");
      if (env_var && strlen(env_var) < 1000)
	strcat(strcpy(pathbuf, env_var), "/scheme");
      else
	strcpy(pathbuf, "/usr/local/bin/scheme");
      execlp(pathbuf, "scheme", "-interactive", "-load", "scheme/startup.scm",
	     (char *)0);
      perror(pathbuf);
      _exit(-1);
    }
  else
    {
      /* I am the parent process.  Return with good fd's */
      scheme_pid = pid_my_child;
      filedes[0] = pipe_child_to_parent[0];
      filedes[1] = pipe_parent_to_child[1];
      return 0;
    }
  /*NOTREACHED*/
  return 0;			/* gcc -Wall thinks this is reached */
}

/* The first trick BRL-BUDDY learns is to take input with a separator
 * at the beginning and end and just send the middle, e.g. input
 * "SEP=9595\nHello WorldSEP=9595\n" and output "Hello World"
 */

/*ARGSUSED*/
void
brl_trick_out(input_file, output_file)
     FILE *input_file, output_file; /* my in/output = scheme's out/input */
{
  char separator[BUFSIZ+1], buf[BUFSIZ+1], *ptr;

  separator[BUFSIZ]='\0';
  fgets(separator, BUFSIZ, input_file);
  while (fgets(buf, BUFSIZ, input_file))
    {
      ptr = strstr(buf, separator);
      if (ptr) *ptr = '\0';
      fputs(buf, stdout);
      if (ptr) break;
    }
  return;
}

char *
brl_get_username(buf, buflen)
     char *buf;
     int buflen;
{
  char *ptr1, *ptr2, *ptr3;
  int userlen;
  static char *webuser="webuser";

  ptr1=getenv("SSL_CLIENT_DN");
  if (!ptr1 || !*ptr1)
    {
      ptr1=getenv("REMOTE_USER");
      if (ptr1 && *ptr1) strncpy(buf, ptr1, buflen);
      else strncpy(buf, webuser, buflen);
      return buf;
    }
  ptr2=strstr(ptr1, "/Email=");
  if (ptr2) ptr3=strchr(ptr2+7, '@');
  if (ptr2 && ptr3)
    {
      /* /CN=Bruce Lewis/Email=brlewis@MIT.EDU */
      /*                ptr2          ptr3     */
      userlen = ptr3 - (ptr2+7) + 1; /* length of username + 1 char */
      if (buflen > userlen) buflen=userlen;
      strncpy(buf, ptr2+7, buflen);
      buf[buflen-1]='\0';
    }
  else
    strncpy(buf, webuser, buflen);
  return buf;
}

#ifdef HAVE_LIBCLIENT		/* OCI client library */

void
brl_trick_sql(input_file, output_file)
     FILE *input_file, *output_file; /* my in/output = scheme's out/input */
{
  char linebuf[BUFSIZ+1], *sqlbuf;
  char userbuf[256];
  int sqllen;
#ifdef HAVE_LIBFCGI
  char **environ_fcgi;
#endif

  brl_get_username(userbuf, 246);
#ifdef BRL_SUFFIX
  strcat(userbuf, BRL_SUFFIX);
#endif
  fgets(linebuf, BUFSIZ, input_file); /* FIXME: honor protocol request */
  fgets(linebuf, BUFSIZ, input_file); /* get length of sql statement */
  sqlbuf=malloc(1+(sqllen=atoi(linebuf)));
  if (!sqlbuf)
    {
      fputs(scm_nomem, output_file);
      return;
    }
  fread(sqlbuf, 1, sqllen, input_file);
  sqlbuf[sqllen]='\0';

#ifdef HAVE_LIBFCGI
  environ_fcgi = environ;
  environ = environ_initial;	/* prevents FCGI from bashing OCI env vars */
#endif
  brl_oci_dosql(userbuf, sqlbuf, output_file);
#ifdef HAVE_LIBFCGI
  environ = environ_fcgi;
#endif
  return;
}

#endif /* HAVE_LIBCLIENT - OCI client library */

static struct brl_trick_t
{
  char *tname;
  void (*tproc)();
  int tcontinue;
} brl_tricks[] =
  {
    { "out", brl_trick_out, 0 },
#ifdef HAVE_LIBCLIENT		/* OCI client library */
    { "sql", brl_trick_sql, 1 },
#endif /* HAVE_LIBCLIENT - OCI client library */
    { 0, 0 }
  };

void main ()
{
#ifdef HAVE_LIBCLIENT
  FILE *ociconnect;
#endif
  FILE *output_from_scheme, *input_to_scheme, *real_input_to_scheme;
  int scheme_fd[2], retval, i;
  char scheme_buf[BUFSIZ+1], *path_translated, *ptr;
  char b1[BUFSIZ+8], b2[BUFSIZ+8]; /* for setbuf() (Irix recommends +8) */
  static char *part1="(brl-cgi-req2 '";
  static char *part2=" '";
  static char *part3=" ";
  static char *part4=")\n";
  struct sigaction sig_act;
  sigset_t sig_mask;

#ifndef NO_BRL_DEBUG
  char debugtpl[31];
  time_t now;

  debugfd = mkstemp(strcpy(debugtpl, "/usr/tmp/brl-fcgi-XXXXXX"));
#endif

  /* Go to the starting directory */
  ptr = getenv("BRL_HOME");
  if (!ptr) ptr = BRL_DIR_DEFAULT;
  if (chdir(ptr))
    {
      perror(ptr);
      brl_die(0);
    }

  /* Set up handler to kill scheme and die gracefully on various signals */
  sig_act.sa_handler = brl_die;
  sigfillset(&(sig_act.sa_mask));
  sigemptyset(&sig_mask);
  for (i=0; brl_catches[i]; i++)
    {
      sigaction(brl_catches[i], &sig_act, NULL);
      sigaddset(&sig_mask, brl_catches[i]);
    }
  sigprocmask(SIG_UNBLOCK, &sig_mask, 0);

#ifdef HAVE_LIBCLIENT
  if (NULL == (ociconnect = fopen("Private/ociconnect", "r")))
    {
      perror("Private/ociconnect");
      exit(1);
    }
  else
    {
      char conn_string[256], *ptr1;

      /* Read the connection string for the account creator */
      fgets(conn_string, 256, ociconnect);
      /* Trim whitespace */
      for(ptr1=conn_string+strlen(conn_string-1);
	  (ptr1 > conn_string) && isspace(*ptr1);
	  ptr1--)
	*ptr1 = '\0';
      brl_oci_init(conn_string);
      /* Erase secret info */
      memset(conn_string, 0, 256);
      /* Revoke privilege to read the file */
      setuid(getuid());
    }
#endif

  /* initialize variables that don't vary with each request */
  scheme_buf[BUFSIZ]='\0';	/* avoid runaway strings */

  /* Start scheme interpreter */
  retval = scheme_pipe(scheme_fd);
  if (!retval)
    {
      output_from_scheme = fdopen(scheme_fd[0], "r");
      real_input_to_scheme = fdopen(scheme_fd[1], "w");
    }
  /* Big efficiency gain from buffered I/O */
  /* FCGI should have tmpfile() */
  input_to_scheme=fopen(tmpnam(NULL), "w+");
  if (!input_to_scheme) retval = -1;

#ifdef HAVE_LIBFCGI
  environ_initial = environ;

  while(FCGI_Accept() >= 0) {
#endif
    client_connected=1;		/* hello! */
    alarm(55);			/* don't let scheme hang */

    if (retval < 0 || (waitpid(scheme_pid, NULL, WNOHANG) != 0))
      {
	perror("last error");
	brl_die(0);		/* does not return */
      }
    else
      {
	path_translated = getenv("PATH_TRANSLATED");
	if (!path_translated) path_translated="";

	/* Send the scheme command down the pipe */
	fputs(part1, input_to_scheme);
	environ2sexp(input_to_scheme);
	fputs(part2, input_to_scheme);
	query2sexp(input_to_scheme);
	fputs(part3, input_to_scheme);
	brl_str2sexp(input_to_scheme, getenv("PATH_TRANSLATED"), -1);
	fputs(part4, input_to_scheme);
	brl_flush(input_to_scheme, real_input_to_scheme);

	while (1)
	  {
	    if (fgets(scheme_buf, BUFSIZ, output_from_scheme))
	      {
#ifndef NO_BRL_DEBUG
		if (debugfd != -1) write(debugfd, scheme_buf, strlen(scheme_buf));
#endif
		if (!strncmp(scheme_buf, ";Aborting!", 10))
		  {
		    /* ignore-errors choked, e.g. recursion depth */
		    printf("Content-type: text/html\r\n\r\n");
		    printf("<h1>Scheme error</h1>\n");
		    printf("<samp>%s</samp><p>\n", scheme_buf);
		    printf("<!-- %s -->\n", rcsid);
		    break;
		  }
		if (!strncmp(scheme_buf, "BRL-BUDDY: ", 11))
		  {
		    /* scheme wants me to do something */
		    /* trim trailing space from command */
		    for(ptr=scheme_buf+strlen(scheme_buf-1);
			!isalnum(*ptr); ptr--) *ptr = '\0';
		    ptr=scheme_buf+11;
		    for(i=0; brl_tricks[i].tname; i++)
		      {
			if (!strcmp(ptr, brl_tricks[i].tname))
			  {
			    brl_tricks[i].tproc(output_from_scheme,	/* my in */
						input_to_scheme); /* my out */
			    brl_flush(input_to_scheme, real_input_to_scheme);
			    break;
			  }
		      }
		    if (!brl_tricks[i].tcontinue) break;
		  }
	      }
	  }
      }
    client_connected=0;		/* bye! */
    alarm(0);			/* cancel alarm */
#ifdef HAVE_LIBFCGI
  } /* while */
#endif
  exit(0);
}
