
; On startup we're connected to our buddy, brlfcgi
(define brl-input-from-buddy (current-input-port))
(define brl-output-to-buddy (current-output-port))

(define (brl-sql-eval-string str)
  (if (null? str)
      '()
      (let* ((thunk (lambda ()
		      (display "\nBRL-BUDDY: sql\n")
		      (display "\n")
		      (display (string-length str)) (display "\n")
		      (display str) (display "\n"))))
	(with-output-to-port brl-output-to-buddy thunk)
	(with-input-from-port brl-input-from-buddy read))))

(define (brl-sql-env x env)
  (let ((result (brl-sql-eval-string
		   (apply string-append (map brl-string x)))))
      (cond ((pair? result)
	     (cond ((eq? 'error (car result)) (apply error (cdr result)))
		   (else (brl-make-table-env result env))))
	    (else result))))

(define brl-sql
  (lambda x
    (let ((result (brl-sql-eval-string
		   (apply string-append (map brl-string x)))))
      (cond ((pair? result)
	     (cond ((eq? 'error (car result)) (apply error (cdr result)))
		   (else (brl-make-table result))))
	    (else result)))))


(define (brl-sql-chl chl)
  (cond ((null? chl) '())
	((equal? #\' (car chl))
	 (cons #\' (cons #\' (brl-sql-chl (cdr chl)))))
	(else (cons (car chl) (brl-sql-chl (cdr chl))))))

(define (brl-sql-string str)
  (list->string
   (append (list #\')
	   (brl-sql-chl (string->list (brl-string str)))
	   (list #\'))))

; Make a date suitable for use with HTTP.  Note the date must be GMT.

(define (brl-sql-http-date datespec)
  (string-append
   "INITCAP(TO_CHAR(" datespec ", 'DY, DD MON YYYY HH24:MM:SS GMT'))"))

; Make a string to update the contents of a table
; (brl-sql-update-string
;   "tablename"
;   (list "name"  "phone")
;   (list "bruce" "'555-1212'")
;   (list "bruce" "'936-1234'")
;   "where id = 55")   ==> "update tablename set phone='936-1234' where id=55"

(define (brl-sql-update-item colname oldval newval)
  (if (equal? oldval newval)
      ""
      (string-append (brl-string colname) "=" (brl-string newval))))

(define (brl-sql-update-items colnames oldvals newvals)
  (brl-nonblanks (map brl-sql-update-item colnames oldvals newvals)))

(define (brl-sql-update-string tablename colnames oldvals newvals wclause)
  (let ((setclauses (brl-separate
		     ", "
		     (brl-nonblanks
		      (map brl-sql-update-item colnames oldvals newvals)))))
    (if (null? setclauses)
	'()
	(string-append
	 "update " tablename "\n  set " setclauses "\n  " wclause))))

(define brl-sql-update
  (lambda x
    (brl-sql-eval-string
     (brl-sql-update-string
      (car x) (cadr x) (caddr x) (cadddr x)
      (apply string-append (map brl-string (cddddr x)))))))

(define (brl-all-equal? l)
  (cond ((or (null? l) (null? (cdr l))) #t)
	((not (equal? (car l) (cadr l))) #f)
	(else (brl-all-equal? (cdr l)))))

(define (brl-collapse-list l)
  (if (brl-all-equal? l)
      (car l)
      l))
;
; Collapse several rows according to identical members.
; This is useful in certain SQL joins.
; (brl-collapse-rows '(
;                      (1 2 4)
;                      (1 2 5))  ==> (1 2 (4 5))
;

(define (brl-collapse-rows r)
  (map brl-collapse-list (apply map (cons list r))))

;
; (brl-equivalence-classes '() '(1 1 1 2 2 2 3 3 3 4) equal?)
;   ==> ((1 1 1) (2 2 2) (3 3 3) (4))
; The list must be sorted so that equivalent items are adjacent.
;

(define (brl-equivalence-classes processed unprocessed same?)
  (if (null? unprocessed)
      (map reverse (reverse processed))
      (if (null? processed)
	  (brl-equivalence-classes (list (list (car unprocessed)))
				   (cdr unprocessed)
				   same?)
	  (if (same? (car unprocessed) (car (car processed)))
	      (brl-equivalence-classes (cons (cons (car unprocessed)
						   (car processed))
					     (cdr processed))
				       (cdr unprocessed)
				       same?)
	      (brl-equivalence-classes (cons (list (car unprocessed))
					     processed)
				       (cdr unprocessed)
				       same?)))))

(define (brl-collapse-contents c ref)
  (define (same? r1 r2) (equal? (list-ref r1 ref) (list-ref r2 ref)))
  (cons (car c)
	(map brl-collapse-rows
	     (brl-equivalence-classes '() (cdr c) same?))))

(define (brl-column-ordinal colname l n)
  (if (null? l)
      (error "No such column: " colname)
      (if (string-ci=? colname (car l))
	  n
	  (brl-column-ordinal colname (cdr l) (+ 1 n)))))
  
(define (brl-collapse-by column-name table)
  (define c (cadr (brl-table-contents table)))
  (brl-make-table-env
   (brl-collapse-contents c (brl-column-ordinal column-name (car c) 0))
   (cadr (brl-table-parent table))))
